from __future__ import annotations

from functools import partial
from itertools import count
from math import e, log, pi, sin, sqrt
from typing import Iterable

from vsexprtools import ExprOp, ExprVars, combine, norm_expr
from vstools import (
    ColorRange, ConvMode, CustomIntEnum, CustomNotImplementedError, FuncExceptT, FunctionUtil, KwargsT, PlanesT,
    StrList, VSFunction, check_ref_clip, flatten_vnodes, get_y, scale_value, vs
)

from .blur import box_blur, gauss_blur

__all__ = [
    'replace_low_frequencies',

    'MeanMode',

    'lehmer_diff_merge'
]


def replace_low_frequencies(
    flt: vs.VideoNode, ref: vs.VideoNode, LFR: float, DCTFlicker: bool = False,
    planes: PlanesT = None, mode: ConvMode = ConvMode.SQUARE
) -> vs.VideoNode:
    func = FunctionUtil(flt, replace_low_frequencies, planes, (vs.YUV, vs.GRAY))

    check_ref_clip(flt, ref, func.func)

    ref_work_clip = get_y(ref) if func.luma_only else ref

    LFR = max(LFR or (300 * func.work_clip.width / 1920), 50.0)

    # Frequency sample rate is resolution * 2 (for Nyquist)
    freq_sample = max(func.work_clip.width, func.work_clip.height) * 2

    k = sqrt(log(2) / 2) * LFR          # Constant for -3dB
    f_cut = freq_sample / (k * 2 * pi)  # Frequency Cutoff for Gaussian Sigma

    expr = 'x y - z + '

    if DCTFlicker:
        sec = scale_value(sin(e) + 0.1, 8, func.work_clip, range_out=ColorRange.FULL)
        expr += f'y z - D! y z = swap dup D@ 0 = 0 D@ 0 < -1 1 ? ? {sec} * + ?'

    flt_blur = gauss_blur(func.work_clip, f_cut, None, mode)
    ref_blur = gauss_blur(ref_work_clip, f_cut, None, mode)

    final = norm_expr([func.work_clip, flt_blur, ref_blur], expr, planes, force_akarin=func.func)

    return func.return_clip(final)


class MeanMode(CustomIntEnum):
    MINIMUM = -2
    HARMONIC = -1
    GEOMETRIC = 0

    ARITHMETIC = 1

    RMS = 2
    CUBIC = 3
    MAXIMUM = 4

    LEHMER = 10

    MINIMUM_ABS = 20
    MAXIMUM_ABS = 21

    def __call__(
        self, *_clips: vs.VideoNode | Iterable[vs.VideoNode], planes: PlanesT = None, func: FuncExceptT | None = None
    ) -> vs.VideoNode:
        func = func or self.__class__

        clips = flatten_vnodes(_clips)

        n_clips = len(clips)
        n_op = n_clips - 1

        if n_clips < 2:
            return next(iter(clips))

        kwargs = KwargsT(planes=planes, func=func)

        if self == MeanMode.MINIMUM:
            return ExprOp.MIN(clips, **kwargs)

        if self == MeanMode.MAXIMUM:
            return ExprOp.MAX(clips, **kwargs)

        if self == MeanMode.GEOMETRIC:
            return combine(clips, ExprOp.MUL, None, None, [1 / n_clips, ExprOp.POW], **kwargs)

        if self == MeanMode.LEHMER:
            counts = range(n_clips)
            clip_vars = ExprVars(n_clips)

            expr = StrList([[f'{clip} range_diff - D{i}!' for i, clip in zip(counts, clip_vars)]])

            for y in range(2):
                expr.extend([
                    [f'D{i}@ {3 - y} pow' for i in counts],
                    ExprOp.ADD * n_op, f'P{y + 1}!'
                ])

            expr.append('P2@ 0 = 0 P1@ P2@ / ? range_diff +')

            return norm_expr(clips, expr, **kwargs)

        if self in {MeanMode.RMS, MeanMode.ARITHMETIC, MeanMode.CUBIC, MeanMode.HARMONIC}:
            return combine(
                clips, ExprOp.ADD, f'{self.value} {ExprOp.POW}', None, [
                    n_clips, ExprOp.DIV, 1 / self, ExprOp.POW
                ], **kwargs
            )

        if self in {MeanMode.MINIMUM_ABS, MeanMode.MAXIMUM_ABS}:
            operator = ExprOp.MIN if self is MeanMode.MINIMUM_ABS else ExprOp.MAX

            expr_string = ''
            for src in ExprVars(n_clips):
                expr_string += f'{src} range_diff - abs {src.upper()}D! '

            for i, src, srcn in zip(count(), ExprVars(n_clips), ExprVars(1, n_clips)):
                expr_string += f'{src.upper()}D@ {srcn.upper()}D@ {operator} {src} '

                if i == n_clips - 2:
                    expr_string += f'{srcn} '

            expr_string += '? ' * n_op

            return norm_expr(clips, expr_string, **kwargs)

        raise CustomNotImplementedError


def lehmer_diff_merge(
    *_clips: vs.VideoNode | Iterable[vs.VideoNode],
    filter: VSFunction | list[VSFunction] = partial(box_blur, radius=3, passes=2),
    planes: PlanesT = None
) -> vs.VideoNode:
    import warnings

    from vsdenoise import frequency_merge

    warnings.warn('lehmer_diff_merge: This function was moved to vsdenoise.frequency_merge, with better output!')

    return frequency_merge(*_clips, tr=0, lowpass=filter, planes=planes)



from __future__ import annotations

from vsexprtools import ExprVars, complexpr_available, norm_expr
from vstools import (
    CustomIndexError, CustomValueError, PlanesT, check_ref_clip, check_variable, core, get_neutral_value,
    get_peak_value, normalize_planes, vs
)

from .enum import LimitFilterMode

__all__ = [
    'limit_filter'
]


def limit_filter(
    flt: vs.VideoNode, src: vs.VideoNode, ref: vs.VideoNode | None = None,
    mode: LimitFilterMode = LimitFilterMode.CLAMPING, planes: PlanesT = None,
    thr: int | tuple[int, int] = 1, elast: float = 2.0, bright_thr: int | None = None
) -> vs.VideoNode:
    assert check_variable(src, limit_filter)
    assert check_variable(flt, limit_filter)
    check_ref_clip(src, flt, limit_filter)
    check_ref_clip(flt, ref, limit_filter)

    if ref is not None:
        assert check_variable(ref, limit_filter)

    planes = normalize_planes(flt, planes)

    is_yuv = flt.format.color_family == vs.YUV

    got_ref = ref is not None

    if isinstance(thr, tuple):
        thr, thrc = thr
    else:
        thrc = thr

    if bright_thr is None:
        bright_thr = thr

    for var, name, l in [
        (thr, 'thr', 0), (thrc, 'thrc', 0), (bright_thr, 'bright_thr', 0), (elast, 'elast', 1)
    ]:
        if var < l:
            raise CustomIndexError(f'{name} must be >= {l}', limit_filter, reason=var)

    if ref is None and mode != LimitFilterMode.CLAMPING:
        raise CustomValueError('You need to specify ref!', limit_filter, reason='mode={mode}', mode=mode)

    force_expr = mode.force_expr

    if any([
        got_ref, flt.format.sample_type == vs.FLOAT,
        thr >= 128, bright_thr >= 128, mode != LimitFilterMode.CLAMPING
    ]):
        force_expr = True

    if thr <= 0 and bright_thr <= 0 and (not is_yuv or thrc <= 0):
        return src

    if thr >= 255 and bright_thr >= 255 and (not is_yuv or thrc >= 255):
        return flt

    if force_expr:
        peak = get_peak_value(flt)

        clips = [flt, src]

        if ref:
            clips.append(ref)

        return norm_expr(clips, (
            _limit_filter_expr(got_ref, thr, elast, bright_thr, peak, mode),
            _limit_filter_expr(got_ref, thrc, elast, thrc, peak, mode)
        ))

    diff = flt.std.MakeDiff(src, planes)

    diff = _limit_filter_lut(diff, elast, thr, bright_thr, [0])

    if 1 in planes or 2 in planes:
        diff = _limit_filter_lut(diff, elast, thrc, thrc, list({*planes} - {0}))

    return flt.std.MakeDiff(diff, planes)


def _limit_filter_lut(
    diff: vs.VideoNode, elast: float, thr: float, largen_thr: float, planes: list[int]
) -> vs.VideoNode:
    assert check_variable(diff, limit_filter)

    neutral = int(get_neutral_value(diff))
    peak = int(get_peak_value(diff))

    thr = int(thr * peak / 255)
    largen_thr = int(largen_thr * peak / 255)

    if thr >= peak / 2 and largen_thr >= peak / 2:
        neutral_clip = diff.std.BlankClip(color=neutral)

        all_planes = list(range(diff.format.num_planes))

        if planes == all_planes:
            return neutral_clip

        diff_planes = planes + list({*all_planes} - {*planes})

        return core.std.ShufflePlanes(
            [neutral_clip, diff], diff_planes, diff.format.color_family
        )

    no_elast = elast <= 1

    def limitLut(x: int) -> int:
        dif = x - neutral

        dif_abs = abs(dif)

        thr_1 = largen_thr if dif > 0 else thr

        if dif_abs <= thr_1:
            return neutral

        if no_elast:
            return x

        thr_2 = thr_1 * elast

        if dif_abs >= thr_2:
            return x

        thr_slope = 1 / (thr_2 - thr_1)

        return round(dif * (dif_abs - thr_1) * thr_slope + neutral)

    return diff.std.Lut(planes, function=limitLut)


def _limit_filter_expr(
    got_ref: bool, thr: float, elast: float, largen_thr: float, peak: float, mode: LimitFilterMode
) -> str:
    if mode in {LimitFilterMode.SIMPLE_MIN, LimitFilterMode.SIMPLE_MAX}:
        return f'y z - abs y x - abs {mode.op} z x ?'
    elif mode in {LimitFilterMode.SIMPLE2_MIN, LimitFilterMode.SIMPLE2_MAX}:
        return f'y x - abs z x - abs {mode.op} y z ?'
    elif mode in {LimitFilterMode.DIFF_MIN, LimitFilterMode.DIFF_MAX}:
        if complexpr_available:
            return f'y x - A! y z - B! A@ B@ xor y A@ abs B@ abs {mode.op} x z ? ?'

        return f'y x - y z - xor y y x - abs y z - abs {mode.op} x z ? ?'

    ref = ExprVars[1 + got_ref]

    header = ''

    dif = 'x y -'
    dif_abs = f' x {ref} - abs'

    if complexpr_available:
        header = f'{dif} DIF! {dif_abs} DIFABS!'
        dif, dif_abs = 'DIF@', 'DIFABS@'

    thr, largen_thr = [x * peak / 255 for x in (thr, largen_thr)]

    if thr <= 0 and largen_thr <= 0:
        return 'y'

    if thr >= peak and largen_thr >= peak:
        return ''

    def _limit_xthr_expr(var: float) -> str:
        if var <= 0:
            return 'y'

        if var >= peak:
            return 'x'

        if elast <= 1:
            return f'{dif_abs} {var} <= x y ?'

        thr_1, thr_2 = var, var * elast
        thr_slope = 1 / (thr_2 - thr_1)

        return f'{dif_abs} {thr_1} <= x {dif_abs} {thr_2} >= y y {dif} {thr_2} {dif_abs} - * {thr_slope} * + ? ?'

    limitExpr = _limit_xthr_expr(thr)

    if largen_thr != thr:
        limitExpr = f'x {ref} > {_limit_xthr_expr(largen_thr)} {limitExpr} ?'

    return f'{header} {limitExpr}'



"""Wrapper for RGVS, RGSF, and various other functions"""

__version__ = '1.7.0'

__author_name__, __author_email__ = 'Irrational Encoding Wizardry', 'wizards@encode.moe'
__maintainer_name__, __maintainer_email__ = 'Setsugen no ao', 'setsugen@setsugen.dev'

__author__ = f'{__author_name__} <{__author_email__}>'
__maintainer__ = __author__

if __name__ == '__github__':
    print(__version__)



from __future__ import annotations

from vsexprtools import complexpr_available, expr_func
from vstools import NotFoundEnumValue, PlanesT, check_variable, core, normalize_seq, pick_func_stype, vs

from .aka_expr import (
    aka_removegrain_expr_11_12, aka_removegrain_expr_19, aka_removegrain_expr_20, aka_removegrain_expr_23,
    aka_removegrain_expr_24, removegrain_aka_exprs, repair_aka_exprs
)
from .enum import BlurMatrix, RemoveGrainMode, RemoveGrainModeT, RepairMode, RepairModeT, VerticalCleanerModeT

__all__ = [
    'repair', 'removegrain',
    'clense', 'backward_clense', 'forward_clense',
    'vertical_cleaner', 'horizontal_cleaner'
]


def repair(clip: vs.VideoNode, repairclip: vs.VideoNode, mode: RepairModeT) -> vs.VideoNode:
    assert check_variable(clip, repair)
    assert check_variable(repairclip, repair)

    is_float = clip.format.sample_type == vs.FLOAT
    mode = normalize_seq(mode, clip.format.num_planes)

    if not sum(mode):
        return clip

    if not complexpr_available:
        if (RepairMode.CLIP_REF_RG20 in mode or RepairMode.CLIP_REF_RG23 in mode) and is_float:
            raise NotFoundEnumValue(
                'Specified RepairMode for rgsf is not implemented!', repair, reason=iter(mode)
            )

        return pick_func_stype(clip, core.rgvs.Repair, core.rgsf.Repair)(clip, repairclip, mode)

    return core.akarin.Expr(
        [clip, repairclip], [repair_aka_exprs[m]() for m in mode], clip.format.id, True
    )


def removegrain(clip: vs.VideoNode, mode: RemoveGrainModeT) -> vs.VideoNode:
    assert check_variable(clip, removegrain)

    mode = normalize_seq(mode, clip.format.num_planes)
    mode = list(map(RemoveGrainMode, mode))

    if not sum(mode):
        return clip

    # rgvs is faster for integer clips
    if hasattr(core, 'rgvs') and clip.format.sample_type == vs.INTEGER and all(m in range(24 + 1) for m in mode):
        return clip.rgvs.RemoveGrain(mode)

    if not complexpr_available:
        return clip.rgsf.RemoveGrain(mode)

    expr = list[str]()
    for idx, m in enumerate(mode):
        if m == RemoveGrainMode.SQUARE_BLUR:
            if all(mm == m for mm in mode):
                return BlurMatrix.WMEAN(clip)
            expr.append(aka_removegrain_expr_11_12())
        elif RemoveGrainMode.BOB_TOP_CLOSE <= m <= RemoveGrainMode.BOB_BOTTOM_INTER:
            return pick_func_stype(clip, clip.rgvs.RemoveGrain, clip.rgsf.RemoveGrain)(mode)
        elif m == RemoveGrainMode.CIRCLE_BLUR:
            if set(mode) == {RemoveGrainMode.CIRCLE_BLUR}:
                return BlurMatrix.CIRCLE(clip)
            expr.append(aka_removegrain_expr_19())
        elif m == RemoveGrainMode.BOX_BLUR:
            if set(mode) == {RemoveGrainMode.BOX_BLUR}:
                return BlurMatrix.MEAN(clip)
            expr.append(aka_removegrain_expr_20())
        elif m == RemoveGrainMode.EDGE_DEHALO:
            expr.append(aka_removegrain_expr_23(0 if idx == 0 else -0.5))
        elif m == RemoveGrainMode.EDGE_DEHALO2:
            expr.append(aka_removegrain_expr_24(0 if idx == 0 else -0.5))
        else:
            expr.append(removegrain_aka_exprs[m]())

    return expr_func(clip, expr, opt=True)


def clense(
    clip: vs.VideoNode,
    previous_clip: vs.VideoNode | None = None, next_clip: vs.VideoNode | None = None,
    planes: PlanesT = None
) -> vs.VideoNode:
    return pick_func_stype(clip, clip.rgvs.Clense, clip.rgsf.Clense)(previous_clip, next_clip, planes)


def forward_clense(clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
    return pick_func_stype(clip, clip.rgvs.ForwardClense, clip.rgsf.ForwardClense)(planes)


def backward_clense(clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
    return pick_func_stype(clip, clip.rgvs.BackwardClense, clip.rgsf.BackwardClense)(planes)


def vertical_cleaner(clip: vs.VideoNode, mode: VerticalCleanerModeT) -> vs.VideoNode:
    return pick_func_stype(clip, clip.rgvs.VerticalCleaner, clip.rgsf.VerticalCleaner)(mode)


def horizontal_cleaner(clip: vs.VideoNode, mode: VerticalCleanerModeT) -> vs.VideoNode:
    return vertical_cleaner(clip.std.Transpose(), mode).std.Transpose()



from __future__ import annotations

from typing import Iterable

from vsexprtools import ExprOp, ExprVars, complexpr_available, norm_expr
from vstools import (
    CustomIndexError, CustomOverflowError, PlanesT, VSFunction, check_variable, core, fallback, flatten,
    get_neutral_value, normalize_planes, vs
)

from .enum import LimitFilterMode
from .limit import limit_filter

__all__ = [
    'minimum_diff', 'median_diff',
    'median_clips',
    'flux_smooth'
]


def minimum_diff(
    clip: vs.VideoNode,
    clip_func: VSFunction, diff_func: VSFunction | None = None,
    diff: bool | None = None, planes: PlanesT = None
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    diff = fallback(diff, diff_func is None)

    diff_func = fallback(diff_func, clip_func)

    if not diff:
        return median_clips(clip, clip_func(clip), diff_func(clip), planes=planes)

    filtered = clip_func(clip)

    diffa = core.std.MakeDiff(clip, filtered, planes)

    diffb = diff_func(diffa)

    return median_diff(clip, diffa, diffb, planes)


def median_diff(clip: vs.VideoNode, diffa: vs.VideoNode, diffb: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
    assert check_variable(clip, median_diff)

    planes = normalize_planes(clip, planes)
    neutral = [get_neutral_value(clip), get_neutral_value(clip, True)]

    if complexpr_available:
        expr = 'y z - D! x D@ y {mid} clamp D@ {mid} y clamp - -'
    else:
        expr = 'x y z - y min {mid} max y z - {mid} min y max - -'

    return norm_expr([clip, diffa, diffb], expr, planes, mid=neutral)


def median_clips(*_clips: vs.VideoNode | Iterable[vs.VideoNode], planes: PlanesT = None) -> vs.VideoNode:
    clips = list[vs.VideoNode](flatten(_clips))  # type: ignore
    n_clips = len(clips)

    if not complexpr_available and n_clips > 26:
        raise CustomOverflowError('You can pass only up to 26 clips without akarin Expr!', median_clips, reason=n_clips)
    elif n_clips < 3:
        raise CustomIndexError('You must pass at least 3 clips!', median_clips, reason=n_clips)

    if n_clips == 3:
        return norm_expr(clips, 'x y z min max y z max min')

    all_clips = str(ExprVars(1, n_clips))

    n_ops = n_clips - 2

    yzmin, yzmax = [
        all_clips + f' {op}' * n_ops for op in (ExprOp.MIN, ExprOp.MAX)
    ]

    header = ''
    if complexpr_available:
        header = f'{yzmin} YZMIN! {yzmax} YZMAX! '
        yzmin, yzmax = 'YZMIN@', 'YZMAX@'

    expr = f'{header} x {yzmin} min x = {yzmin} x {yzmax} max x = {yzmax} x ? ?'

    return norm_expr(clips, expr, planes)


def flux_smooth(
    clip: vs.VideoNode, radius: int = 2, threshold: int = 7, scenechange: int = 24, planes: PlanesT = None
) -> vs.VideoNode:
    assert check_variable(clip, flux_smooth)

    if radius < 1 or radius > 7:
        raise CustomIndexError('Radius must be between 1 and 7 (inclusive)!', flux_smooth, reason=radius)

    planes = normalize_planes(clip, planes)

    threshold = threshold << clip.format.bits_per_sample - 8

    cthreshold = threshold if (1 in planes or 2 in planes) else 0

    median = clip.tmedian.TemporalMedian(radius, planes)  # type: ignore
    average = clip.focus2.TemporalSoften2(  # type: ignore
        radius, threshold, cthreshold, scenechange
    )

    return limit_filter(average, clip, median, LimitFilterMode.DIFF_MIN, planes)



from functools import partial
from typing import Callable

from ._rg import *  # noqa: F401, F403
from ._rg import (  # noqa: F401
    aka_removegrain_expr_1, aka_removegrain_expr_2_4, aka_removegrain_expr_5, aka_removegrain_expr_6,
    aka_removegrain_expr_7, aka_removegrain_expr_8, aka_removegrain_expr_9, aka_removegrain_expr_10,
    aka_removegrain_expr_11_12, aka_removegrain_expr_17, aka_removegrain_expr_18, aka_removegrain_expr_19,
    aka_removegrain_expr_20, aka_removegrain_expr_21_22, aka_removegrain_expr_23, aka_removegrain_expr_24,
    aka_removegrain_expr_26, aka_removegrain_expr_27, aka_removegrain_expr_28
)
from ._rp import *  # noqa: F401, F403
from ._rp import (
    aka_repair_expr_1_4, aka_repair_expr_5, aka_repair_expr_6, aka_repair_expr_7, aka_repair_expr_8, aka_repair_expr_9,
    aka_repair_expr_10, aka_repair_expr_11_14, aka_repair_expr_15, aka_repair_expr_16, aka_repair_expr_17,
    aka_repair_expr_18, aka_repair_expr_19, aka_repair_expr_20, aka_repair_expr_21, aka_repair_expr_22,
    aka_repair_expr_23, aka_repair_expr_24, aka_repair_expr_26, aka_repair_expr_27, aka_repair_expr_28
)


def _noop_expr() -> str:
    return ''


removegrain_aka_exprs = list[Callable[[], str]]([
    _noop_expr, aka_removegrain_expr_1, partial(aka_removegrain_expr_2_4, 2),
    partial(aka_removegrain_expr_2_4, 3), partial(aka_removegrain_expr_2_4, 4),
    aka_removegrain_expr_5, aka_removegrain_expr_6, aka_removegrain_expr_7, aka_removegrain_expr_8,
    aka_removegrain_expr_9, aka_removegrain_expr_10, aka_removegrain_expr_11_12, aka_removegrain_expr_11_12,
    _noop_expr, _noop_expr, _noop_expr, _noop_expr,
    aka_removegrain_expr_17, aka_removegrain_expr_18, aka_removegrain_expr_19, aka_removegrain_expr_20,
    aka_removegrain_expr_21_22, aka_removegrain_expr_21_22, _noop_expr, _noop_expr,
    _noop_expr, aka_removegrain_expr_26, aka_removegrain_expr_27, aka_removegrain_expr_28
])

repair_aka_exprs = list[Callable[[], str]]([
    _noop_expr, partial(aka_repair_expr_1_4, 1), partial(aka_repair_expr_1_4, 2),
    partial(aka_repair_expr_1_4, 3), partial(aka_repair_expr_1_4, 4),
    aka_repair_expr_5, aka_repair_expr_6, aka_repair_expr_7, aka_repair_expr_8,
    aka_repair_expr_9, aka_repair_expr_10, partial(aka_repair_expr_11_14, 1), partial(aka_repair_expr_11_14, 2),
    partial(aka_repair_expr_11_14, 3), partial(aka_repair_expr_11_14, 4), aka_repair_expr_15, aka_repair_expr_16,
    aka_repair_expr_17, aka_repair_expr_18, aka_repair_expr_19, aka_repair_expr_20,
    aka_repair_expr_21, aka_repair_expr_22, aka_repair_expr_23, aka_repair_expr_24,
    _noop_expr, aka_repair_expr_26, aka_repair_expr_27, aka_repair_expr_28
])



from __future__ import annotations

from functools import partial
from inspect import Signature
from typing import Callable

from vsexprtools import complexpr_available, norm_expr
from vstools import (
    CustomValueError, GenericVSFunction, PlanesT, check_ref_clip, check_variable, clamp_arr, get_neutral_value, iterate,
    normalize_planes, to_arr, vs
)

from .blur import blur, box_blur, median_blur, min_blur
from .enum import BlurMatrix, RemoveGrainMode, RemoveGrainModeT, RepairMode, RepairModeT
from .rgtools import removegrain, repair
from .util import norm_rmode_planes

__all__ = [
    'contrasharpening', 'contra',
    'contrasharpening_dehalo', 'contra_dehalo',
    'contrasharpening_median', 'contra_median',
    'fine_contra'
]


def contrasharpening(
    flt: vs.VideoNode, src: vs.VideoNode, radius: int | list[int] = 1,
    mode: RepairModeT = RepairMode.MINMAX_SQUARE3, planes: PlanesT = 0
) -> vs.VideoNode:
    """
    contra-sharpening: sharpen the denoised clip, but don't add more to any pixel than what was previously removed.
    Script by Didée, at the VERY GRAINY thread (http://forum.doom9.org/showthread.php?p=1076491#post1076491)
    :param flt:         Filtered clip
    :param src:         Source clip
    :param radius:      Spatial radius for contra-sharpening (1-3). Default is 2 for HD / 1 for SD.
    :param mode:        Mode of rgvs.Repair to limit the difference
    :param planes:      Planes to process, defaults to None
    :return:            Contrasharpened clip
    """

    assert check_variable(src, contrasharpening)
    assert check_variable(flt, contrasharpening)
    check_ref_clip(src, flt, contrasharpening)

    if flt.format.sample_type == vs.INTEGER:
        neutral = [get_neutral_value(flt), get_neutral_value(flt, True)]
    else:
        neutral = [0.0]

    planes = normalize_planes(flt, planes)

    # Damp down remaining spots of the denoised clip
    mblur = min_blur(flt, radius, planes)

    rg11 = blur(mblur, radius, planes=planes)

    # Difference of a simple kernel blur
    diff_blur = mblur.std.MakeDiff(rg11, planes)

    # Difference achieved by the filtering
    diff_flt = src.std.MakeDiff(flt, planes)

    # Limit the difference to the max of what the filtering removed locally
    limit = repair(diff_blur, diff_flt, norm_rmode_planes(flt, mode, planes))

    # abs(diff) after limiting may not be bigger than before
    # Apply the limited difference (sharpening is just inverse blurring)
    if complexpr_available:
        expr = 'x {mid} - LD! y {mid} - BD! LD@ abs BD@ abs < LD@ BD@ ? z +'
    else:
        expr = 'x {mid} - abs y {mid} - abs < x y ? {mid} - z +'

    return norm_expr([limit, diff_blur, flt], expr, planes, mid=neutral)


def contrasharpening_dehalo(
    flt: vs.VideoNode, src: vs.VideoNode, level: float = 1.4, alpha: float = 2.49, planes: PlanesT = 0
) -> vs.VideoNode:
    """
    :param dehaloed:    Dehaloed clip
    :param src:         Source clip
    :param level:       Strength level
    :return:            Contrasharpened clip
    """
    assert check_variable(src, contrasharpening)
    assert check_variable(flt, contrasharpening)
    check_ref_clip(src, flt, contrasharpening)

    planes = normalize_planes(flt, planes)

    rep_modes = norm_rmode_planes(flt, RepairMode.MINMAX_SQUARE1, planes)

    weighted = BlurMatrix.WMEAN(flt, planes)
    weighted2 = median_blur(weighted, 2, planes=planes)
    weighted2 = iterate(weighted2, partial(repair, repairclip=weighted), 2, mode=rep_modes)

    neutral = [get_neutral_value(flt), get_neutral_value(flt, True)]

    if complexpr_available:
        clips = [weighted, weighted2, src, flt]
        expr = f'x y - {alpha} * {level} * D! z a - DY! D@ DY@ * 0 < 0 D@ abs DY@ abs < D@ DY@ ? ? a +'
    else:
        diff = norm_expr([weighted, weighted2], f'x y - {alpha} * {level} * {{mid}} +', planes, mid=neutral)

        clips = [diff, src, flt]
        expr = 'x {mid} - y z - xor 0 x {mid} - abs y z - abs < x {mid} - y z - ? ? z +'

    return norm_expr(clips, expr, planes, mid=neutral)


def contrasharpening_median(
    flt: vs.VideoNode, src: vs.VideoNode,
    mode: RemoveGrainModeT | Callable[..., vs.VideoNode] = box_blur,
    planes: PlanesT = 0
) -> vs.VideoNode:
    """
    :param flt:         Filtered clip
    :param src:         Source clip
    :param mode:        Function or the RemoveGrain mode used to blur/repair the filtered clip.
    :param planes:      Planes to process, defaults to None
    :return:            Contrasharpened clip
    """
    assert check_variable(src, contrasharpening)
    assert check_variable(flt, contrasharpening)
    check_ref_clip(src, flt, contrasharpening)

    planes = normalize_planes(flt, planes)

    if isinstance(mode, (int, list, RemoveGrainMode)):
        repaired = removegrain(flt, norm_rmode_planes(flt, mode, planes))
    elif callable(mode):
        repaired = mode(flt, planes=planes)
    else:
        raise CustomValueError('Invalid mode or function passed!', contrasharpening_median)

    if complexpr_available:
        expr = 'x dup + z - D! x y < D@ x y clamp D@ y x clamp ?'
    else:
        expr = 'x dup + z - x y min max x y max min'

    return norm_expr([flt, src, repaired], expr, planes)


def fine_contra(
    flt: vs.VideoNode, src: vs.VideoNode, sharp: float | list[float] | range = 0.75,
    radius: int | list[int] = 1, merge_func: GenericVSFunction | None = None,
    mode: RepairModeT = RepairMode.MINMAX_SQUARE_REF3, planes: PlanesT = 0
) -> vs.VideoNode:
    """
    :param flt:         Filtered clip.
    :param src:         Source clip.
    :param sharp:       Contrast Adaptive Sharpening's sharpening strength.
                        If it's a list, depending on ``merge_func`` being ``None``,
                        it will iterate over with different strengths or merge all with ``merge_func``.
    :param radius:      Spatial radius for contra-sharpening (1-3). Default is 2 for HD / 1 for SD.
    :param merge_func:  Depending on ``sharp``, this will get all sharpened clips and merge them.
    :param mode:        Mode of rgvs.Repair to limit the difference.
    :param planes:      Planes to process, defaults to None.
    :return:            Contrasharpened clip.
    """

    assert check_variable(src, contrasharpening)
    assert check_variable(flt, contrasharpening)
    check_ref_clip(src, flt, contrasharpening)

    if flt.format.sample_type == vs.INTEGER:
        neutral = [get_neutral_value(flt), get_neutral_value(flt, True)]
    else:
        neutral = [0.0]

    planes = normalize_planes(flt, planes)

    mblur = min_blur(flt, radius, planes)

    sharp = [1.0 / x for x in sharp if x] if isinstance(sharp, range) else to_arr(sharp)
    sharp = clamp_arr(sharp, 0.0, 1.0)

    if merge_func is None:
        for s in sharp:
            mblur = mblur.cas.CAS(s, planes)
    else:
        mblurs = [mblur.cas.CAS(s, planes) for s in sharp]

        got_p = 'planes' in Signature.from_callable(merge_func).parameters.keys()

        try:
            if got_p:
                mblur = merge_func(*mblurs, planes=planes)
            else:
                mblur = merge_func(*mblurs)
        except Exception:
            if got_p:
                mblur = merge_func(mblurs, planes=planes)
            else:
                mblur = merge_func(mblurs)

    limit = repair(mblur, src.std.MakeDiff(flt, planes), norm_rmode_planes(flt, mode, planes))

    if complexpr_available:
        expr = 'x {mid} - LD! y {mid} - BD! LD@ abs BD@ abs < LD@ BD@ ? z +'
    else:
        expr = 'x {mid} - abs y {mid} - abs < x y ? {mid} - z +'

    return norm_expr([limit, mblur, flt], expr, planes, mid=neutral)


contra = contrasharpening
contra_dehalo = contrasharpening_dehalo
contra_median = contrasharpening_median



# flake8: noqa
from . import aka_expr, util

from .blur import *
from .contra import *
from .enum import *
from .freqs import *
from .func import *
from .limit import *
from .rgtools import *
from .sharp import *



A1 = 'x[-1,-1]'
A2 = 'x[0,-1]'
A3 = 'x[1,-1]'
A4 = 'x[-1,0]'
A5 = 'x[1,0]'
A6 = 'x[-1,1]'
A7 = 'x[0,1]'
A8 = 'x[1,1]'
c = 'x'
PIXELS = ' '.join([A1, A2, A3, A4, A5, A6, A7, A8])


def aka_removegrain_expr_1() -> str:
    return f'x {PIXELS} min min min min min min min {PIXELS} max max max max max max max clamp'


def aka_removegrain_expr_2_4(m: int) -> str:
    # print(f'dup{9 - m} dup{m - 1}')
    return f'{PIXELS} sort8 dup{8 - m} max_val! dup{m - 1} min_val! drop8 x min_val@ max_val@ clamp'


def aka_removegrain_expr_5() -> str:
    return (
        f'x {A1} {A8} min {A1} {A8} max clamp clamp1! '
        f'x {A2} {A7} min {A2} {A7} max clamp clamp2! '
        f'x {A3} {A6} min {A3} {A6} max clamp clamp3! '
        f'x {A4} {A5} min {A4} {A5} max clamp clamp4! '
        'x clamp1@ - abs c1! '
        'x clamp2@ - abs c2! '
        'x clamp3@ - abs c3! '
        'x clamp4@ - abs c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_removegrain_expr_6() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        'x clamp1@ - abs 2 * d1@ + c1! '
        'x clamp2@ - abs 2 * d2@ + c2! '
        'x clamp3@ - abs 2 * d3@ + c3! '
        'x clamp4@ - abs 2 * d4@ + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_removegrain_expr_7() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        # Only change is removing the "* 2"
        'x clamp1@ - abs d1@ + c1! '
        'x clamp2@ - abs d2@ + c2! '
        'x clamp3@ - abs d3@ + c3! '
        'x clamp4@ - abs d4@ + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_removegrain_expr_8() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        'x clamp1@ - abs d1@ 2 * + c1! '
        'x clamp2@ - abs d2@ 2 * + c2! '
        'x clamp3@ - abs d3@ 2 * + c3! '
        'x clamp4@ - abs d4@ 2 * + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_removegrain_expr_9() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'd1@ d2@ d3@ d4@ min min min mindiff! '
        'mindiff@ d4@ = x mil4@ mal4@ clamp mindiff@ d2@ = x mil2@ mal2@ clamp '
        'mindiff@ d3@ = x mil3@ mal3@ clamp x mil1@ mal1@ clamp ? ? ?'
    )


def aka_removegrain_expr_10() -> str:
    return (
        f'x {A1} - abs d1! '
        f'x {A2} - abs d2! '
        f'x {A3} - abs d3! '
        f'x {A4} - abs d4! '
        f'x {A5} - abs d5! '
        f'x {A6} - abs d6! '
        f'x {A7} - abs d7! '
        f'x {A8} - abs d8! '
        'd1@ d2@ d3@ d4@ d5@ d6@ d7@ d8@ min min min min min min min mindiff! '
        f'mindiff@ d7@ = {A7} mindiff@ d8@ = {A8} mindiff@ d6@ = {A6} mindiff@ d2@ = {A2} '
        f'mindiff@ d3@ = {A3} mindiff@ d1@ = {A1} mindiff@ d5@ = {A5} {A4} ? ? ? ? ? ? ?'
    )


def aka_removegrain_expr_11_12() -> str:
    return f'x 4 * {A2} {A4} {A5} {A7} + + + 2 * + {A1} {A3} {A6} {A8} + + + + 16 /'


def aka_removegrain_expr_17() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        'x maxmil@ minmal@ min maxmil@ minmal@ max clamp'
    )


def aka_removegrain_expr_18() -> str:
    return (
        f'x {A1} - abs x {A8} - abs max d1! '
        f'x {A2} - abs x {A7} - abs max d2! '
        f'x {A3} - abs x {A6} - abs max d3! '
        f'x {A4} - abs x {A5} - abs max d4! '
        'd1@ d2@ d3@ d4@ min min min mindiff! '
        f'mindiff@ d4@ = x {A4} {A5} min {A4} {A5} max clamp '
        f'mindiff@ d2@ = x {A2} {A7} min {A2} {A7} max clamp '
        f'mindiff@ d3@ = x {A3} {A6} min {A3} {A6} max clamp '
        f'x {A1} {A8} min {A1} {A8} max clamp ? ? ?'
    )


def aka_removegrain_expr_19() -> str:
    return f'{A1} {A2} {A3} {A4} {A5} {A6} {A7} {A8} + + + + + + + 8.0 /'


def aka_removegrain_expr_20() -> str:
    return f'x {A1} {A2} {A3} {A4} {A5} {A6} {A7} {A8} + + + + + + + + 9.0 /'


def aka_removegrain_expr_21_22() -> str:
    return (
        f'{A1} {A8} + 2 / av1! '
        f'{A2} {A7} + 2 / av2! '
        f'{A3} {A6} + 2 / av3! '
        f'{A4} {A5} + 2 / av4! '
        'x av1@ av2@ av3@ av4@ min min min av1@ av2@ av3@ av4@ max max max clamp'
    )


def aka_removegrain_expr_23(peak_min: float) -> str:
    minmax = f'min max max max {peak_min} max'
    u = f'x mal1@ - linediff1@ min x mal2@ - linediff2@ min x mal3@ - linediff3@ min x mal4@ - linediff4@ {minmax}'
    d = f'mil1@ x - linediff1@ min mil2@ x - linediff2@ min mil3@ x - linediff3@ min mil4@ x - linediff4@ {minmax}'
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - linediff1! '
        'mal2@ mil2@ - linediff2! '
        'mal3@ mil3@ - linediff3! '
        'mal4@ mil4@ - linediff4! '
        f'x {u} - {d} +'
    )


def aka_removegrain_expr_24(peak_min: float) -> str:
    linediff_minmax = (
        'linediff1@ t1@ - t1@ min linediff2@ t2@ - t2@ min '
        'linediff3@ t3@ - t3@ min linediff4@ t4@ - t4@ min '
        f'max max max {peak_min} max'
    )

    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - linediff1! '
        'mal2@ mil2@ - linediff2! '
        'mal3@ mil3@ - linediff3! '
        'mal4@ mil4@ - linediff4! '
        'x mal1@ - t1! '
        'x mal2@ - t2! '
        'x mal3@ - t3! '
        'x mal4@ - t4! '
        f'{linediff_minmax} u! '
        'mil1@ x - t1! '
        'mil2@ x - t2! '
        'mil3@ x - t3! '
        'mil4@ x - t4! '
        f'{linediff_minmax} d! '
        f'x u@ - d@ +'
    )


def aka_removegrain_expr_25() -> str:
    raise NotImplementedError


def aka_removegrain_expr_26() -> str:
    return (
        f'{A1} {A2} min mil1! '
        f'{A1} {A2} max mal1! '
        f'{A2} {A3} min mil2! '
        f'{A2} {A3} max mal2! '
        f'{A3} {A5} min mil3! '
        f'{A3} {A5} max mal3! '
        f'{A5} {A8} min mil4! '
        f'{A5} {A8} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A7} {A8} min mil1! '
        f'{A7} {A8} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A4} {A6} min mil3! '
        f'{A4} {A6} max mal3! '
        f'{A1} {A4} min mil4! '
        f'{A1} {A4} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x maxmil@ minmal@ min maxmil@ minmal@ max clamp'
    )


def aka_removegrain_expr_27() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A1} {A2} min mil2! '
        f'{A1} {A2} max mal2! '
        f'{A7} {A8} min mil3! '
        f'{A7} {A8} max mal3! '
        f'{A2} {A7} min mil4! '
        f'{A2} {A7} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A2} {A3} min mil1! '
        f'{A2} {A3} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A3} {A5} min mil4! '
        f'{A3} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        f'{A4} {A6} min mil1! '
        f'{A4} {A6} max mal1! '
        f'{A4} {A5} min mil2! '
        f'{A4} {A5} max mal2! '
        f'{A5} {A8} min mil3! '
        f'{A5} {A8} max mal3! '
        f'{A1} {A4} min mil4! '
        f'{A1} {A4} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x maxmil@ minmal@ min maxmil@ minmal@ max clamp'
    )


def aka_removegrain_expr_28() -> str:
    return (
        f'{A1} {A2} min mil1! '
        f'{A1} {A2} max mal1! '
        f'{A2} {A3} min mil2! '
        f'{A2} {A3} max mal2! '
        f'{A3} {A5} min mil3! '
        f'{A3} {A5} max mal3! '
        f'{A5} {A8} min mil4! '
        f'{A5} {A8} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A7} {A8} min mil1! '
        f'{A7} {A8} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A4} {A6} min mil3! '
        f'{A4} {A6} max mal3! '
        f'{A1} {A5} min mil4! '
        f'{A1} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A3} {A6} min mil2! '
        f'{A3} {A6} max mal2! '
        f'{A2} {A7} min mil3! '
        f'{A2} {A7} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x maxmil@ minmal@ min maxmil@ minmal@ max clamp'
    )



from __future__ import annotations

from typing import Any, Sequence, TypeVar, cast

from vstools import (
    GenericVSFunction, KwargsT, Nb, PlanesT, check_variable, check_variable_format, join, normalize_planes,
    normalize_seq, plane, vs
)

from .enum import RemoveGrainMode, RepairMode, BlurMatrix

__all__ = [
    'wmean_matrix', 'mean_matrix',
    'norm_rmode_planes',
    'normalize_radius'
]

wmean_matrix = list(BlurMatrix.WMEAN)
mean_matrix = list(BlurMatrix.MEAN)

RModeT = TypeVar('RModeT', RemoveGrainMode, RepairMode)


def norm_rmode_planes(
    clip: vs.VideoNode, mode: int | RModeT | Sequence[int | RModeT], planes: PlanesT = None
) -> list[int]:
    assert check_variable(clip, norm_rmode_planes)

    modes_array = normalize_seq(mode, clip.format.num_planes)

    planes = normalize_planes(clip, planes)

    return [
        cast(RModeT, rep if i in planes else 0) for i, rep in enumerate(modes_array, 0)
    ]


def normalize_radius(
    clip: vs.VideoNode, func: GenericVSFunction, radius: list[Nb] | tuple[str, list[Nb]],
    planes: list[int], **kwargs: Any
) -> vs.VideoNode:
    assert check_variable_format(clip, normalize_radius)

    name, radius = radius if isinstance(radius, tuple) else ('radius', radius)

    radius = normalize_seq(radius, clip.format.num_planes)

    def _get_kwargs(rad: Nb) -> KwargsT:
        return kwargs | {name: rad, 'planes': planes}

    if len(set(radius)) > 0:
        if len(planes) != 1:
            return join([
                func(plane(clip, i), **_get_kwargs(rad)) for i, rad in enumerate(radius)
            ])

        radius_i = radius[planes[0]]
    else:
        radius_i = radius[0]

    return func(clip, **_get_kwargs(radius_i))



from __future__ import annotations

from vsexprtools import norm_expr
from vstools import CustomTypeError, PlanesT, VSFunction, check_ref_clip, check_variable, normalize_planes, vs

from .blur import gauss_blur, min_blur
from .enum import BlurMatrix
from .limit import limit_filter
from .util import normalize_radius

__all__ = [
    'unsharpen',
    'unsharp_masked',
    'limit_usm',
    'shimmer_soothe'
]


def unsharpen(
    clip: vs.VideoNode, strength: float = 1.0, sigma: float | list[float] = 1.5,
    prefilter: vs.VideoNode | VSFunction | None = None
) -> vs.VideoNode:
    assert check_variable(clip, unsharpen)

    ref = prefilter(clip) if callable(prefilter) else prefilter

    check_ref_clip(clip, ref)

    den = ref or clip
    blur = gauss_blur(den, sigma)

    unsharp = norm_expr([den, blur], f'x y - {strength} * x +', 0)

    if ref is not None:
        return unsharp

    return unsharp.std.MergeDiff(clip.std.MakeDiff(den))


def unsharp_masked(
    clip: vs.VideoNode, radius: int | list[int] = 1, strength: float = 100.0, planes: PlanesT = None
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, unsharp_masked, radius, planes, strength=strength)

    blurred = BlurMatrix.LOG(radius, strength)(clip, planes)

    return norm_expr([clip, blurred], 'x dup y - +')


def limit_usm(
    clip: vs.VideoNode, blur: int | vs.VideoNode | VSFunction = 1, thr: int | tuple[int, int] = 3,
    elast: float = 4.0, bright_thr: int | None = None, planes: PlanesT = None
) -> vs.VideoNode:
    """Limited unsharp_masked."""

    if callable(blur):
        blurred = blur(clip)
    elif isinstance(blur, vs.VideoNode):
        blurred = blur  # type: ignore
    elif blur <= 0:  # type: ignore
        blurred = min_blur(clip, -blur, planes)  # type: ignore
    elif blur == 1:
        blurred = BlurMatrix.WMEAN(clip, planes)
    elif blur == 2:
        blurred = BlurMatrix.MEAN(clip, planes)
    else:
        raise CustomTypeError("'blur' must be an int, clip or a blurring function!", limit_usm, blur)

    sharp = norm_expr([clip, blurred], 'x dup y - +', planes)

    return limit_filter(sharp, clip, thr=thr, elast=elast, bright_thr=bright_thr)


def shimmer_soothe(
    sharp: vs.VideoNode, orig: vs.VideoNode, strength: int = 75, tr: int = 1,
    scenechange: bool = False, planes: PlanesT = 0
) -> vs.VideoNode:
    diff = orig.std.MakeDiff(sharp, planes)

    strength = 100 - max(min(100, strength), 0)

    soft_diff = diff.std.AverageFrames([1] * (tr * 2 + 1), None, scenechange, planes)

    limit_diff = norm_expr([diff, soft_diff], [
        'x range_diff - y range_diff - xor x range_diff - 100 / {strength} '
        f'* range_diff + x range_diff - abs y range_diff - abs > x {strength} '
        f'* y 100 {strength} - * + 100 / x ? ?'
    ], planes, strength=strength)

    return norm_expr([orig, limit_diff], 'x y range_diff - -', planes)



from __future__ import annotations

from enum import auto
from math import ceil, exp, log2, pi, sqrt
from typing import Generic, Sequence, TypeVar

from vstools import ConvMode, CustomEnum, CustomIntEnum, CustomValueError, Nb, PlanesT, core, iterate, to_singleton, vs

__all__ = [
    'LimitFilterMode',
    'RemoveGrainMode', 'RemoveGrainModeT',
    'RepairMode', 'RepairModeT',
    'VerticalCleanerMode', 'VerticalCleanerModeT',
    'BlurMatrix'
]


class LimitFilterModeMeta:
    force_expr = True


class LimitFilterMode(LimitFilterModeMeta, CustomIntEnum):
    """Two sources, one filtered"""
    SIMPLE_MIN = auto()
    SIMPLE_MAX = auto()
    """One source, two filtered"""
    SIMPLE2_MIN = auto()
    SIMPLE2_MAX = auto()
    DIFF_MIN = auto()
    DIFF_MAX = auto()
    """One/Two sources, one filtered"""
    CLAMPING = auto()

    @property
    def op(self) -> str:
        return '<' if 'MIN' in self._name_ else '>'

    def __call__(self, force_expr: bool = True) -> LimitFilterMode:
        self.force_expr = force_expr

        return self


class RemoveGrainMode(CustomIntEnum):
    NONE = 0
    MINMAX_AROUND1 = 1
    MINMAX_AROUND2 = 2
    MINMAX_AROUND3 = 3
    MINMAX_MEDIAN = 4
    EDGE_CLIP_STRONG = 5
    EDGE_CLIP_MODERATE = 6
    EDGE_CLIP_MEDIUM = 7
    EDGE_CLIP_LIGHT = 8
    LINE_CLIP_CLOSE = 9
    MIN_SHARP = 10
    SQUARE_BLUR = 11
    BOB_TOP_CLOSE = 13
    BOB_BOTTOM_CLOSE = 14
    BOB_TOP_INTER = 15
    BOB_BOTTOM_INTER = 16
    MINMAX_MEDIAN_OPP = 17
    LINE_CLIP_OPP = 18
    CIRCLE_BLUR = 19
    BOX_BLUR = 20
    OPP_CLIP_AVG = 21
    OPP_CLIP_AVG_FAST = 22
    EDGE_DEHALO = 23
    EDGE_DEHALO2 = 24
    MIN_SHARP2 = 25
    SMART_RGC = 26
    SMART_RGCL = 27
    SMART_RGCL2 = 28

    def __call__(self, clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import removegrain
        from .util import norm_rmode_planes
        return removegrain(clip, norm_rmode_planes(clip, self, planes))


RemoveGrainModeT = int | RemoveGrainMode | Sequence[int | RemoveGrainMode]


class RepairMode(CustomIntEnum):
    NONE = 0
    MINMAX_SQUARE1 = 1
    MINMAX_SQUARE2 = 2
    MINMAX_SQUARE3 = 3
    MINMAX_SQUARE4 = 4
    LINE_CLIP_MIN = 5
    LINE_CLIP_LIGHT = 6
    LINE_CLIP_MEDIUM = 7
    LINE_CLIP_STRONG = 8
    LINE_CLIP_CLOSE = 9
    MINMAX_SQUARE_REF_CLOSE = 10
    MINMAX_SQUARE_REF1 = 11
    MINMAX_SQUARE_REF2 = 12
    MINMAX_SQUARE_REF3 = 13
    MINMAX_SQUARE_REF4 = 14
    CLIP_REF_RG5 = 15
    CLIP_REF_RG6 = 16
    CLIP_REF_RG17 = 17
    CLIP_REF_RG18 = 18
    CLIP_REF_RG19 = 19
    CLIP_REF_RG20 = 20
    CLIP_REF_RG21 = 21
    CLIP_REF_RG22 = 22
    CLIP_REF_RG23 = 23
    CLIP_REF_RG24 = 24
    CLIP_REF_RG26 = 26
    CLIP_REF_RG27 = 27
    CLIP_REF_RG28 = 28

    def __call__(self, clip: vs.VideoNode, repairclip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import repair
        from .util import norm_rmode_planes
        return repair(clip, repairclip, norm_rmode_planes(clip, self, planes))


RepairModeT = int | RepairMode | Sequence[int | RepairMode]


class VerticalCleanerMode(CustomIntEnum):
    NONE = 0
    MEDIAN = 1
    PRESERVING = 2

    def __call__(self, clip: vs.VideoNode, planes: PlanesT = None) -> vs.VideoNode:
        from .rgtools import vertical_cleaner
        from .util import norm_rmode_planes
        return vertical_cleaner(clip, norm_rmode_planes(clip, self, planes))


VerticalCleanerModeT = int | VerticalCleanerMode | Sequence[int | VerticalCleanerMode]


class BaseBlurMatrix(Generic[Nb], list[Nb]):
    def __call__(
        self, clip: vs.VideoNode, planes: PlanesT = None, mode: ConvMode = ConvMode.SQUARE,
        bias: float | None = None, divisor: float | None = None, saturate: int | None = None,
        passes: int = 1
    ) -> vs.VideoNode:
        if len(self) <= 1:
            return clip

        conv = self.tosizes(*range(3, 26, 2))

        return iterate(clip, core.std.Convolution, passes, conv, bias, divisor, planes, saturate, mode)

    @property
    def asint(self) -> BaseBlurMatrix[int]:
        return BaseBlurMatrix[int](map(int, self))

    @property
    def asfloat(self) -> BaseBlurMatrix[float]:
        return BaseBlurMatrix[float](map(float, self))

    def tosize(self: BBMatrixT, size: int) -> BBMatrixT:
        curlen = len(self)

        diff = curlen - size

        if diff == 0:
            return self

        lh = abs(diff) // 2
        rh = abs(diff) - lh

        if diff < 0:
            return self.__class__(self[:lh] + self + self[-rh:])

        return self.__class__(self[lh:-rh])

    def tosizes(self: BBMatrixT, *_sizes: int) -> BBMatrixT:
        sizes = list(sorted(_sizes, reverse=True))

        curlen = len(self)

        for sizeh, sizel in zip(sizes, sizes[1:]):
            if curlen >= sizeh:
                return self.tosize(sizeh)

            if curlen == sizel:
                return self.tosize(sizel)

            if curlen > sizel:
                if (sizeh - curlen) < (curlen - sizel):
                    return self.tosize(sizeh)

                return self.tosize(sizel)

        return self.tosize(sizes[-1])


BBMatrixT = TypeVar('BBMatrixT', bound=BaseBlurMatrix)  # type: ignore


class BlurMatrix(BaseBlurMatrix[int], CustomEnum):  # type: ignore
    BOX = [1, 1, 0, 1, 1, 0, 0, 0, 0]
    MEAN = [1, 1, 1, 1, 1, 1, 1, 1, 1]
    WMEAN = [1, 2, 1, 2, 4, 2, 1, 2, 1]
    CIRCLE = [1, 1, 1, 1, 0, 1, 1, 1, 1]

    @to_singleton.as_property
    class GAUSS:
        def __call__(self, sigma: float, taps: int | None = None, scale_value: float = 1023) -> BaseBlurMatrix[float]:
            taps = self.get_taps(sigma, taps)

            if taps < 0:
                raise CustomValueError('Taps must be >= 0!')

            if sigma > 0.0:
                half_pisqrt = 1.0 / sqrt(2.0 * pi) * sigma
                doub_qsigma = 2 * sigma ** 2

                high, *kernel = [half_pisqrt * exp(-x ** 2 / doub_qsigma) for x in range(taps + 1)]

                kernel = [x * scale_value / high for x in kernel]
                kernel = [*kernel[::-1], scale_value, *kernel]
            else:
                kernel = [scale_value]

            return BaseBlurMatrix[float](kernel)

        def from_radius(self, radius: int) -> BaseBlurMatrix[float]:
            return self((radius + 1.0) / 3)

        def get_taps(self, sigma: float, taps: int | None = None) -> int:
            if taps is None:
                taps = ceil(abs(sigma) * 8 + 1) // 2

            return taps

    @to_singleton.as_property
    class LOG:
        def __call__(self, radius: int = 1, strength: float = 100.0) -> BaseBlurMatrix[float]:
            strength = max(1e-6, min(log2(3) * strength / 100, log2(3)))

            weight = 0.5 ** strength / ((1 - 0.5 ** strength) * 0.5)

            matrix = [1.0]

            for _ in range(radius):
                matrix.append(matrix[-1] / weight)

            kernel = [*matrix[::-1], *matrix[1:]]

            return BaseBlurMatrix[float](kernel)



from __future__ import annotations

from functools import lru_cache, partial
from itertools import count
from math import e, log2
from typing import Any

from vsexprtools import ExprOp, ExprVars, complexpr_available, norm_expr
from vskernels import Bilinear, Gaussian
from vspyplugin import FilterMode, ProcessMode, PyPluginCuda
from vstools import (
    ConvMode, CustomNotImplementedError, CustomRuntimeError, FunctionUtil, NotFoundEnumValue, PlanesT, StrList,
    check_variable, core, depth, fallback, get_depth, get_neutral_value, join, normalize_planes, normalize_seq, split,
    to_arr, vs
)

from .enum import BlurMatrix, LimitFilterMode
from .limit import limit_filter
from .util import normalize_radius

__all__ = [
    'blur', 'box_blur', 'side_box_blur',
    'gauss_blur',
    'min_blur', 'sbr', 'median_blur',
    'bilateral'
]


def blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, mode: ConvMode = ConvMode.SQUARE, planes: PlanesT = None
) -> vs.VideoNode:
    assert check_variable(clip, blur)

    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, blur, radius, planes, mode=mode)

    if mode == ConvMode.SQUARE:
        matrix2 = [1, 3, 4, 3, 1]
        matrix3 = [1, 4, 8, 10, 8, 4, 1]
    elif mode in {ConvMode.HORIZONTAL, ConvMode.VERTICAL}:
        matrix2 = [1, 6, 15, 20, 15, 6, 1]
        matrix3 = [1, 10, 45, 120, 210, 252, 210, 120, 45, 10, 1]
    else:
        raise NotFoundEnumValue('Invalid mode specified!', blur)

    if radius == 1:
        matrix = [1, 2, 1]
    elif radius == 2:
        matrix = matrix2
    elif radius == 3:
        matrix = matrix3
    else:
        raise CustomNotImplementedError('This radius isn\'t supported!', blur)

    return clip.std.Convolution(matrix, planes=planes, mode=mode)


def box_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, passes: int = 1,
    mode: ConvMode = ConvMode.SQUARE, planes: PlanesT = None
) -> vs.VideoNode:
    assert check_variable(clip, box_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, box_blur, radius, planes, passes=passes)

    if not radius:
        return clip

    fp16 = clip.format.sample_type == vs.FLOAT and clip.format.bits_per_sample == 16

    if radius > 12 and not fp16:
        blurred = clip.std.BoxBlur(
            planes,
            radius, 0 if mode == ConvMode.VERTICAL else passes,
            radius, 0 if mode == ConvMode.HORIZONTAL else passes
        )
    else:
        matrix_size = radius * 2 | 1

        if fp16:
            matrix_size **= 2

        blurred = clip
        for _ in range(passes):
            if fp16:
                blurred = norm_expr(blurred, [
                    ExprOp.matrix('x', radius, mode=mode), ExprOp.ADD * (matrix_size - 1), matrix_size, ExprOp.DIV
                ], planes)
            else:
                blurred = blurred.std.Convolution([1] * matrix_size, planes=planes, mode=mode)

    return blurred


def side_box_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, planes: PlanesT = None,
    inverse: bool = False
) -> vs.VideoNode:
    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, side_box_blur, radius, planes, inverse=inverse)

    half_kernel = [(1 if i <= 0 else 0) for i in range(-radius, radius + 1)]

    conv_m1 = partial(core.std.Convolution, matrix=half_kernel, planes=planes)
    conv_m2 = partial(core.std.Convolution, matrix=half_kernel[::-1], planes=planes)
    blur_pt = partial(core.std.BoxBlur, planes=planes)

    vrt_filters, hrz_filters = [
        [
            partial(conv_m1, mode=mode), partial(conv_m2, mode=mode),
            partial(blur_pt, hradius=hr, vradius=vr, hpasses=h, vpasses=v)
        ] for h, hr, v, vr, mode in [
            (0, None, 1, radius, ConvMode.VERTICAL), (1, radius, 0, None, ConvMode.HORIZONTAL)
        ]
    ]

    vrt_intermediates = (vrt_flt(clip) for vrt_flt in vrt_filters)
    intermediates = list(
        hrz_flt(vrt_intermediate)
        for i, vrt_intermediate in enumerate(vrt_intermediates)
        for j, hrz_flt in enumerate(hrz_filters) if not i == j == 2
    )

    comp_blur = None if inverse else box_blur(clip, radius, 1, planes)

    if complexpr_available:
        template = '{cum} x - abs {new} x - abs < {cum} {new} ?'

        cum_expr, cumc = '', 'y'
        n_inter = len(intermediates)

        for i, newc, var in zip(count(), ExprVars[2:26], ExprVars[4:26]):
            if i == n_inter - 1:
                break

            cum_expr += template.format(cum=cumc, new=newc)

            if i != n_inter - 2:
                cumc = var.upper()
                cum_expr += f' {cumc}! '
                cumc = f'{cumc}@'

        if comp_blur:
            clips = [clip, *intermediates, comp_blur]
            cum_expr = f'x {cum_expr} - {ExprVars[n_inter + 1]} +'
        else:
            clips = [clip, *intermediates]

        cum = norm_expr(clips, cum_expr, planes, force_akarin='vsrgtools.side_box_blur')
    else:
        cum = intermediates[0]
        for new in intermediates[1:]:
            cum = limit_filter(clip, cum, new, LimitFilterMode.SIMPLE2_MIN, planes)

        if comp_blur:
            cum = clip.std.MakeDiff(cum).std.MergeDiff(comp_blur)

    if comp_blur:
        return box_blur(cum, 1, min(radius // 2, 1))

    return cum


def gauss_blur(
    clip: vs.VideoNode, sigma: float | list[float] = 0.5, taps: int | None = None,
    mode: ConvMode = ConvMode.SQUARE, planes: PlanesT = None
) -> vs.VideoNode:
    assert check_variable(clip, gauss_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(sigma, list):
        return normalize_radius(clip, gauss_blur, ('sigma', sigma), planes, mode=mode)

    if mode in ConvMode.VERTICAL:
        sigma = min(sigma, clip.height)

    if mode in ConvMode.HORIZONTAL:
        sigma = min(sigma, clip.width)

    taps = BlurMatrix.GAUSS.get_taps(sigma, (orig_taps := taps))

    no_fmtc = not hasattr(core, 'fmtc')

    kernel = BlurMatrix.GAUSS(sigma, taps, 1.0 if no_fmtc and taps > 12 else 1023)

    if len(kernel) <= 25:
        return kernel(clip, planes, mode)

    if no_fmtc:
        if not complexpr_available:
            raise CustomRuntimeError(
                'With a high sigma you need a high number of taps, '
                'and that\'t only supported with fmtc scaling or akarin expr!'
                '\nInstall one of the two plugins or set lower (<= 12) the taps!'
            )

        proc: vs.VideoNode = clip

        if ConvMode.HORIZONTAL in mode:
            proc = ExprOp.convolution('x', kernel, mode=ConvMode.HORIZONTAL)(proc)

        if ConvMode.VERTICAL in mode:
            proc = ExprOp.convolution('x', kernel, mode=ConvMode.VERTICAL)(proc)

        return proc

    def _fmtc_blur(plane: vs.VideoNode) -> vs.VideoNode:
        wdown, hdown = plane.width, plane.height

        if ConvMode.HORIZONTAL in mode:
            wdown = round(max(round(wdown / sigma), 2) / 2) * 2

        if ConvMode.VERTICAL in mode:
            hdown = round(max(round(hdown / sigma), 2) / 2) * 2

        down = Bilinear.scale(plane, wdown, hdown)

        return Gaussian(curve=9, taps=min(fallback(orig_taps, 30), 128)).scale(down, plane.width, plane.height)

    if not {*range(clip.format.num_planes)} - {*planes}:
        return _fmtc_blur(clip)

    return join([
        _fmtc_blur(p) if i in planes else p
        for i, p in enumerate(split(clip))
    ])


def min_blur(clip: vs.VideoNode, radius: int | list[int] = 1, planes: PlanesT = None) -> vs.VideoNode:
    """
    MinBlur by Didée (http://avisynth.nl/index.php/MinBlur)
    Nifty Gauss/Median combination
    """
    assert check_variable(clip, min_blur)

    planes = normalize_planes(clip, planes)

    if isinstance(radius, list):
        return normalize_radius(clip, min_blur, radius, planes)

    if radius in {0, 1}:
        median = clip.std.Median(planes)
    else:
        median = median_blur(clip, radius, planes=planes)

    if radius:
        weighted = blur(clip, radius)
    else:
        weighted = sbr(clip, planes=planes)

    return limit_filter(weighted, clip, median, LimitFilterMode.DIFF_MIN, planes)


def sbr(
    clip: vs.VideoNode,
    radius: int | list[int] = 1, mode: ConvMode = ConvMode.SQUARE,
    planes: PlanesT = None
) -> vs.VideoNode:
    assert check_variable(clip, sbr)

    planes = normalize_planes(clip, planes)

    neutral = [get_neutral_value(clip), get_neutral_value(clip, True)]

    blur_func = partial(blur, radius=radius, mode=mode, planes=planes)

    weighted = blur_func(clip)

    diff = clip.std.MakeDiff(weighted, planes)

    diff_weighted = blur_func(diff)

    clips = [diff, diff_weighted]

    if complexpr_available:
        clips.append(clip)
        expr = 'x y - A! x {mid} - XD! z A@ XD@ * 0 < {mid} A@ abs XD@ abs < A@ {mid} + x ? ? {mid} - -'
    else:
        expr = 'x y - x {mid} - * 0 < {mid} x y - abs x {mid} - abs < x y - {mid} + x ? ?'

    diff = norm_expr(clips, expr, planes, mid=neutral)

    if complexpr_available:
        return diff

    return clip.std.MakeDiff(diff, planes)


def median_blur(
    clip: vs.VideoNode, radius: int | list[int] = 1, mode: ConvMode = ConvMode.SQUARE, planes: PlanesT = None
) -> vs.VideoNode:
    def _get_vals(radius: int) -> tuple[StrList, int, int, int]:
        matrix = ExprOp.matrix('x', radius, mode, [(0, 0)])
        rb = len(matrix) + 1
        st = rb - 1
        sp = rb // 2 - 1
        dp = st - 2

        return matrix, st, sp, dp

    return norm_expr(clip, (
        f"{matrix} sort{st} swap{sp} min! swap{sp} max! drop{dp} x min@ max@ clip"
        for matrix, st, sp, dp in map(_get_vals, to_arr(radius))
    ), planes, force_akarin=median_blur)


class BilateralFilter(PyPluginCuda[None]):
    cuda_kernel = 'bilateral'
    filter_mode = FilterMode.Parallel

    input_per_plane = True
    output_per_plane = True

    @PyPluginCuda.process(ProcessMode.SingleSrcIPP)
    def _(self, src: BilateralFilter.DT, dst: BilateralFilter.DT, f: vs.VideoFrame, plane: int, n: int) -> None:
        self.kernel.bilateral[plane](src, dst)

    @lru_cache
    def get_kernel_shared_mem(
        self, plane: int, func_name: str, blk_size_w: int, blk_size_h: int, dtype_size: int
    ) -> int:
        return (2 * self.bil_radius[plane] + blk_size_w) * (2 * self.bil_radius[plane] + blk_size_h) * dtype_size

    def __init__(
        self, clip: vs.VideoNode, sigmaS: float | list[float], sigmaR: float | list[float],
        radius: int | list[int] | None, **kwargs: Any
    ) -> None:
        sigmaS, sigmaR = normalize_seq(sigmaS), normalize_seq(sigmaR)

        sigmaS_scaled, sigmaR_scaled = [
            [(-0.5 / (val * val)) * log2(e) for val in vals]
            for vals in (sigmaS, sigmaR)
        ]

        if radius is None:
            radius = [max(1, round(s * 3)) for s in sigmaS]

        self.bil_radius = normalize_seq(radius)

        return super().__init__(
            clip, kernel_planes_kwargs=[
                dict(sigmaS=s, sigmaR=r, radius=rad)
                for s, r, rad in zip(sigmaS_scaled, sigmaR_scaled, self.bil_radius)
            ], **kwargs
        )


def bilateral(
    clip: vs.VideoNode, sigmaS: float | list[float] = 3.0, sigmaR: float | list[float] = 0.02,
    ref: vs.VideoNode | None = None, radius: int | list[int] | None = None,
    device_id: int = 0, num_streams: int | None = None, use_shared_memory: bool = True,
    block_x: int | None = None, block_y: int | None = None, planes: PlanesT = None,
    *, gpu: bool | None = None
) -> vs.VideoNode:
    func = FunctionUtil(clip, bilateral, planes)

    sigmaS, sigmaR = func.norm_seq(sigmaS), func.norm_seq(sigmaR)

    if gpu is not False:
        if not ref and min(sigmaS) < 4 and PyPluginCuda.backend.is_available:
            block_x = fallback(block_x, block_y, 16)
            block_y = fallback(block_y, block_x)

            return BilateralFilter(
                clip, sigmaS, sigmaR, radius,
                kernel_size=(block_x, block_y),
                use_shared_memory=use_shared_memory
            ).invoke()

        try:
            basic_args, new_args = (sigmaS, sigmaR, radius, device_id), (num_streams, use_shared_memory)

            if hasattr(core, 'bilateralgpu_rtc'):
                if 'ref' in core.bilateralgpu_rtc.Bilateral.signature:  # type: ignore
                    return clip.bilateralgpu_rtc.Bilateral(*basic_args, *new_args, block_x, block_y, ref)
                if not ref:
                    return clip.bilateralgpu_rtc.Bilateral(*basic_args, *new_args, block_x, block_y)

            if hasattr(core, 'bilateralgpu'):
                try:
                    if 'ref' in core.bilateralgpu.Bilateral.signature:  # type: ignore
                        return clip.bilateralgpu.Bilateral(*basic_args, *new_args, ref)
                    if not ref:
                        return clip.bilateralgpu.Bilateral(*basic_args, *new_args)
                except vs.Error:
                    if not ref:
                        # Old versions
                        return clip.bilateralgpu.Bilateral(*basic_args)
        except vs.Error as e:
            # has the plugin but no cuda GPU available
            if 'cudaGetDeviceCount' in str(e):
                pass

    if (bits := get_depth(clip)) > 16:
        clip = depth(clip, 16)

    if ref and clip.format != ref.format:
        ref = depth(ref, clip)

    clip = clip.bilateral.Bilateral(ref, sigmaS, sigmaR)

    return depth(clip, bits)



A1 = 'y[-1,-1]'
A2 = 'y[0,-1]'
A3 = 'y[1,-1]'
A4 = 'y[-1,0]'
A5 = 'y[1,0]'
A6 = 'y[-1,1]'
A7 = 'y[0,1]'
A8 = 'y[1,1]'
c = 'y'
val = 'x'
PIXELS = ' '.join([A1, A2, A3, A4, A5, A6, A7, A8])


def aka_repair_expr_1_4(m: int) -> str:
    return f'{PIXELS} y sort9 dup{9 - m} max_val! dup{m - 1} min_val! drop9 x min_val@ max_val@ clamp'


def aka_repair_expr_5() -> str:
    return (
        f'x y {A1} {A8} min min y {A1} {A8} max max clamp clamp1! '
        f'x y {A2} {A7} min min y {A2} {A7} max max clamp clamp2! '
        f'x y {A3} {A6} min min y {A3} {A6} max max clamp clamp3! '
        f'x y {A4} {A5} min min y {A4} {A5} max max clamp clamp4! '
        'x clamp1@ - abs c1! '
        'x clamp2@ - abs c2! '
        'x clamp3@ - abs c3! '
        'x clamp4@ - abs c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_repair_expr_6() -> str:
    return (
        f'y {A1} {A8} min min mil1! '
        f'y {A1} {A8} max max mal1! '
        f'y {A2} {A7} min min mil2! '
        f'y {A2} {A7} max max mal2! '
        f'y {A3} {A6} min min mil3! '
        f'y {A3} {A6} max max mal3! '
        f'y {A4} {A5} min min mil4! '
        f'y {A4} {A5} max max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        'x clamp1@ - abs 2 * d1@ + c1! '
        'x clamp2@ - abs 2 * d2@ + c2! '
        'x clamp3@ - abs 2 * d3@ + c3! '
        'x clamp4@ - abs 2 * d4@ + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_repair_expr_7() -> str:
    return (
        f'y {A1} {A8} min min mil1! '
        f'y {A1} {A8} max max mal1! '
        f'y {A2} {A7} min min mil2! '
        f'y {A2} {A7} max max mal2! '
        f'y {A3} {A6} min min mil3! '
        f'y {A3} {A6} max max mal3! '
        f'y {A4} {A5} min min mil4! '
        f'y {A4} {A5} max max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        # Only change is removing the "* 2"
        'x clamp1@ - abs d1@ + c1! '
        'x clamp2@ - abs d2@ + c2! '
        'x clamp3@ - abs d3@ + c3! '
        'x clamp4@ - abs d4@ + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_repair_expr_8() -> str:
    return (
        f'y {A1} {A8} min min mil1! '
        f'y {A1} {A8} max max mal1! '
        f'y {A2} {A7} min min mil2! '
        f'y {A2} {A7} max max mal2! '
        f'y {A3} {A6} min min mil3! '
        f'y {A3} {A6} max max mal3! '
        f'y {A4} {A5} min min mil4! '
        f'y {A4} {A5} max max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'x mil1@ mal1@ clamp clamp1! '
        'x mil2@ mal2@ clamp clamp2! '
        'x mil3@ mal3@ clamp clamp3! '
        'x mil4@ mal4@ clamp clamp4! '
        'x clamp1@ - abs d1@ 2 * + c1! '
        'x clamp2@ - abs d2@ 2 * + c2! '
        'x clamp3@ - abs d3@ 2 * + c3! '
        'x clamp4@ - abs d4@ 2 * + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = clamp4@ mindiff@ c2@ = clamp2@ mindiff@ c3@ = clamp3@ clamp1@ ? ? ?'
    )


def aka_repair_expr_9() -> str:
    return (
        f'y {A1} {A8} min min mil1! '
        f'y {A1} {A8} max max mal1! '
        f'y {A2} {A7} min min mil2! '
        f'y {A2} {A7} max max mal2! '
        f'y {A3} {A6} min min mil3! '
        f'y {A3} {A6} max max mal3! '
        f'y {A4} {A5} min min mil4! '
        f'y {A4} {A5} max max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'd1@ d2@ d3@ d4@ min min min mindiff! '
        'mindiff@ d4@ = x mil4@ mal4@ clamp mindiff@ d2@ = x mil2@ mal2@ clamp '
        'mindiff@ d3@ = x mil3@ mal3@ clamp x mil1@ mal1@ clamp ? ? ?'
    )


def aka_repair_expr_10() -> str:
    return (
        f'x {A1} - abs d1! '
        f'x {A2} - abs d2! '
        f'x {A3} - abs d3! '
        f'x {A4} - abs d4! '
        f'x {A5} - abs d5! '
        f'x {A6} - abs d6! '
        f'x {A7} - abs d7! '
        f'x {A8} - abs d8! '
        f'x y - abs dc! '
        'd1@ d2@ d3@ d4@ d5@ d6@ d7@ d8@ dc@ min min min min min min min min mindiff! '
        f'mindiff@ d7@ = {A7} mindiff@ d8@ = {A8} mindiff@ d6@ = {A6} mindiff@ d2@ = {A2} '
        f'mindiff@ d3@ = {A3} mindiff@ d1@ = {A1} mindiff@ d5@ = {A5} mindiff@ dc@ = y {A4} ? ? ? ? ? ? ? ?'
    )


def aka_repair_expr_11_14(m: int) -> str:
    return f'{PIXELS} sort8 dup{8 - m} max_val! dup{m - 1} min_val! drop8 x y min_val@ min y max_val@ max clamp'


def aka_repair_expr_15() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'y mil1@ mal1@ clamp clamp1! '
        'y mil2@ mal2@ clamp clamp2! '
        'y mil3@ mal3@ clamp clamp3! '
        'y mil4@ mal4@ clamp clamp4! '
        'y clamp1@ - abs c1! '
        'y clamp2@ - abs c2! '
        'y clamp3@ - abs c3! '
        'y clamp4@ - abs c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = x y mil4@ min y mal4@ max clamp '
        'mindiff@ c2@ = x y mil2@ min y mal2@ max clamp '
        'mindiff@ c3@ = x y mil3@ min y mal3@ max clamp '
        'x y mil1@ min y mal1@ max clamp ? ? ?'
    )


def aka_repair_expr_16() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mal1@ mil1@ - d1! '
        'mal2@ mil2@ - d2! '
        'mal3@ mil3@ - d3! '
        'mal4@ mil4@ - d4! '
        'y y mil1@ mal1@ clamp - abs 2 * d1@ + c1! '
        'y y mil2@ mal2@ clamp - abs 2 * d2@ + c2! '
        'y y mil3@ mal3@ clamp - abs 2 * d3@ + c3! '
        'y y mil4@ mal4@ clamp - abs 2 * d4@ + c4! '
        'c1@ c2@ c3@ c4@ min min min mindiff! '
        'mindiff@ c4@ = x y mil4@ min y mal4@ max clamp '
        'mindiff@ c2@ = x y mil2@ min y mal2@ max clamp '
        'mindiff@ c3@ = x y mil3@ min y mal3@ max clamp '
        'x y mil1@ min y mal1@ max clamp ? ? ?'
    )


def aka_repair_expr_17() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A2} {A7} min mil2! '
        f'{A2} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        'x y maxmil@ minmal@ min min y maxmil@ minmal@ max max clamp'
    )


def aka_repair_expr_18() -> str:
    return (
        f'y {A1} - abs y {A8} - abs max d1! '
        f'y {A2} - abs y {A7} - abs max d2! '
        f'y {A3} - abs y {A6} - abs max d3! '
        f'y {A4} - abs y {A5} - abs max d4! '
        'd1@ d2@ d3@ d4@ min min min mindiff! '
        f'mindiff@ d4@ = x {A4} {A5} min y min {A4} {A5} max y max clamp '
        f'mindiff@ d2@ = x {A2} {A7} min y min {A2} {A7} max y max clamp '
        f'mindiff@ d3@ = x {A3} {A6} min y min {A3} {A6} max y max clamp '
        f'x {A1} {A8} min y min {A1} {A8} max y max clamp ? ? ?'
    )


def aka_repair_expr_19() -> str:
    return (
        f'y {A1} - abs y {A2} - abs y {A3} - abs y {A4} - abs y {A5} - abs y {A6} - abs y {A7} - abs y {A8} - abs '
        'min min min min min min min mindiff! '
        'x y mindiff@ - y mindiff@ + clamp'
    )


def aka_repair_expr_20() -> str:
    return (
        f'y {A1} - abs d1! '
        f'y {A2} - abs d2! '
        f'y {A3} - abs d3! '
        f'y {A4} - abs d4! '
        f'y {A5} - abs d5! '
        f'y {A6} - abs d6! '
        f'y {A7} - abs d7! '
        f'y {A8} - abs d8! '
        'd1@ d2@ min mindiff! '
        'd1@ d2@ max maxdiff! '
        'maxdiff@ mindiff@ d3@ clamp maxdiff! '
        'mindiff@ d3@ min mindiff! '
        'maxdiff@ mindiff@ d4@ clamp maxdiff! '
        'mindiff@ d4@ min mindiff! '
        'maxdiff@ mindiff@ d5@ clamp maxdiff! '
        'mindiff@ d5@ min mindiff! '
        'maxdiff@ mindiff@ d6@ clamp maxdiff! '
        'mindiff@ d6@ min mindiff! '
        'maxdiff@ mindiff@ d7@ clamp maxdiff! '
        'mindiff@ d7@ min mindiff! '
        'maxdiff@ mindiff@ d8@ clamp maxdiff! '
        'x y maxdiff@ - y maxdiff@ + clamp'
    )


def aka_repair_expr_21() -> str:
    return (
        f'{A1} {A8} max y - y {A1} {A8} min - max {A2} {A7} max y - y {A2} {A7} min - max '
        f'{A3} {A6} max y - y {A3} {A6} min - max {A4} {A5} max y - y {A4} {A5} min - max min min min u! '
        'x y u@ - y u@ + clamp'
    )


def aka_repair_expr_22() -> str:
    return (
        f'x {A1} - abs x {A2} - abs x {A3} - abs x {A4} - abs x {A5} - abs x {A6} - abs x {A7} - abs x {A8} - abs '
        'min min min min min min min mindiff! '
        'y x mindiff@ - x mindiff@ + clamp'
    )


def aka_repair_expr_23() -> str:
    return (
        f'x {A1} - abs d1! '
        f'x {A2} - abs d2! '
        f'x {A3} - abs d3! '
        f'x {A4} - abs d4! '
        f'x {A5} - abs d5! '
        f'x {A6} - abs d6! '
        f'x {A7} - abs d7! '
        f'x {A8} - abs d8! '
        'd1@ d2@ min mindiff! '
        'd1@ d2@ max maxdiff! '
        'maxdiff@ mindiff@ d3@ clamp maxdiff! '
        'mindiff@ d3@ min mindiff! '
        'maxdiff@ mindiff@ d4@ clamp maxdiff! '
        'mindiff@ d4@ min mindiff! '
        'maxdiff@ mindiff@ d5@ clamp maxdiff! '
        'mindiff@ d5@ min mindiff! '
        'maxdiff@ mindiff@ d6@ clamp maxdiff! '
        'mindiff@ d6@ min mindiff! '
        'maxdiff@ mindiff@ d7@ clamp maxdiff! '
        'mindiff@ d7@ min mindiff! '
        'maxdiff@ mindiff@ d8@ clamp maxdiff! '
        'y x maxdiff@ - x maxdiff@ + clamp'
    )


def aka_repair_expr_24() -> str:
    return (
        f'{A1} {A8} max x - x {A1} {A8} min - max {A2} {A7} max x - x {A2} {A7} min - max '
        f'{A3} {A6} max x - x {A3} {A6} min - max {A4} {A5} max x - x {A4} {A5} min - max min min min u! '
        'y x u@ - x u@ + clamp'
    )


def aka_repair_expr_26() -> str:
    return (
        f'{A1} {A2} min mil1! '
        f'{A1} {A2} max mal1! '
        f'{A2} {A3} min mil2! '
        f'{A2} {A3} max mal2! '
        f'{A3} {A5} min mil3! '
        f'{A3} {A5} max mal3! '
        f'{A5} {A8} min mil4! '
        f'{A5} {A8} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A7} {A8} min mil1! '
        f'{A7} {A8} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A4} {A6} min mil3! '
        f'{A4} {A6} max mal3! '
        f'{A1} {A4} min mil4! '
        f'{A1} {A4} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x y maxmil@ minmal@ min min y maxmil@ minmal@ max max clamp'
    )


def aka_repair_expr_27() -> str:
    return (
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A1} {A2} min mil2! '
        f'{A1} {A2} max mal2! '
        f'{A7} {A8} min mil3! '
        f'{A7} {A8} max mal3! '
        f'{A2} {A7} min mil4! '
        f'{A2} {A7} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A2} {A3} min mil1! '
        f'{A2} {A3} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A3} {A6} min mil3! '
        f'{A3} {A6} max mal3! '
        f'{A3} {A5} min mil4! '
        f'{A3} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        f'{A4} {A6} min mil1! '
        f'{A4} {A6} max mal1! '
        f'{A4} {A5} min mil2! '
        f'{A4} {A5} max mal2! '
        f'{A5} {A8} min mil3! '
        f'{A5} {A8} max mal3! '
        f'{A1} {A4} min mil4! '
        f'{A1} {A4} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x y maxmil@ minmal@ min min y maxmil@ minmal@ max max clamp'
    )


def aka_repair_expr_28() -> str:
    return (
        f'{A1} {A2} min mil1! '
        f'{A1} {A2} max mal1! '
        f'{A2} {A3} min mil2! '
        f'{A2} {A3} max mal2! '
        f'{A3} {A5} min mil3! '
        f'{A3} {A5} max mal3! '
        f'{A5} {A8} min mil4! '
        f'{A5} {A8} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ min min min minmal! '
        f'{A7} {A8} min mil1! '
        f'{A7} {A8} max mal1! '
        f'{A6} {A7} min mil2! '
        f'{A6} {A7} max mal2! '
        f'{A4} {A6} min mil3! '
        f'{A4} {A6} max mal3! '
        f'{A1} {A5} min mil4! '
        f'{A1} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        f'{A1} {A8} min mil1! '
        f'{A1} {A8} max mal1! '
        f'{A3} {A6} min mil2! '
        f'{A3} {A6} max mal2! '
        f'{A2} {A7} min mil3! '
        f'{A2} {A7} max mal3! '
        f'{A4} {A5} min mil4! '
        f'{A4} {A5} max mal4! '
        'mil1@ mil2@ mil3@ mil4@ maxmil@ max max max max maxmil! '
        'mal1@ mal2@ mal3@ mal4@ minmal@ min min min min minmal! '
        'x y maxmil@ minmal@ min min y maxmil@ minmal@ max max clamp'
    )
